#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

"""This module contains the command line interface"""

import argparse

from aiocache import Cache

from QKDSimkit.Channel import start_channel
from QKDSimkit.Client import start_client
from QKDSimkit.p2p_servers import start_p2p
from QKDSimkit.Server import start_server, start_server_and_channel


def cli():
    """Command line interface
    single entry point for server, client, channel and peer to peer node"""
    parser = argparse.ArgumentParser(description='QKDSimkit interface: choose one of the following programs')
    interfaces = parser.add_subparsers(dest='program')

    #   SERVER PARSER
    #   =============

    server = interfaces.add_parser(name='server', description='Server for QKDSimkit')
    server.add_argument('-a', '--address', default='127.0.0.1:5002', type=str,
                        help='Bind socket to this address (default: %(default)s)')
    channels = server.add_subparsers(title='Channel type', dest='channel_type', required=True,
                                     help='Specify where the channel is (required)')
    parser_l = channels.add_parser('local', help='Run the channel on this machine')
    parser_l.add_argument('-ca', '--channel_address', default=':5000', type=str,
                          help='Bind socket to this address (default: %(default)s)')
    parser_l.add_argument('-n', '--noise', default=0.0, type=float,
                          help='Set a noise value for channel, type a float number in [0,1] (default: %(default)s)')
    parser_l.add_argument('-e', '--eve', action='store_true',
                          help='Add an eavesdropper to the channel')
    parser_e = channels.add_parser('external', help='Connect to an external channel')
    parser_e.add_argument('-ca', '--channel_address', type=str, required=True, help='Address of channel [host:port]')

    #   CLIENT PARSER
    #   =============

    client = interfaces.add_parser(name='client', description='Client for QKDSimkit')
    client.add_argument('alice_address', type=str, help='Address of server/Alice [host:port]')
    client.add_argument('channel_address', type=str, help='Address of channel [host:port]')
    client.add_argument('-n', '--number', default=1, type=int, help="Number of keys (default: %(default)s)")
    client.add_argument('-s', '--size', default=256, type=int, help="Size of keys (default: %(default)s)")

    #   CHANNEL PARSER
    #   ==============

    channel = interfaces.add_parser(name='channel', description='Channel for QKDSimkit')
    channel.add_argument('-a', '--address', default=':5000', type=str,
                         help='Bind socket to this address (default: %(default)s)')
    channel.add_argument('-n', '--noise', default=0.0, type=float,
                         help='Set a noise value for channel, type a float number in [0,1] (default: %(default)s)')
    channel.add_argument('-e', '--eve', action='store_true', help='Add an eavesdropper to the channel')

    #   PEER TO PEER PARSER
    #   ===================

    p2p = interfaces.add_parser(name='p2p', description='Peer to peer interface for QKDSimkit')
    p2p.add_argument('node', choices=['alice', 'bob'],
                     help='Choose how to run this node')
    p2p.add_argument('-c', '--channel_address', default=':5000', type=str,
                     help='Specify the address of the channel [host:port]')
    p2p.add_argument('-a', '--address', default='127.0.0.1:5003', type=str,
                     help='Bind socket to this address (default: %(default)s)')

    args = parser.parse_args()

    if args.program == 'server':
        if args.channel_type == 'local':
            start_server_and_channel(args.channel_address, args.noise, args.eve, args.address)
        if args.channel_type == 'external':
            start_server(args.channel_address, args.address)
    elif args.program == 'client':
        start_client(args.alice_address, args.channel_address, args.number, args.size)
    elif args.program == 'channel':
        start_channel(args.address, args.noise, args.eve)
    elif args.program == 'p2p':
        start_p2p(args.node, args.address, args.channel_address)
    else:
        parser.print_help()


if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

from .alice import *
from .bob import *
from .channel import *
from .node import Node
from .receiver import Receiver
from .sender import Sender
from .utils import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

"""This module contains client's procedure to retrieve keys"""

import argparse
import http.client
import urllib.parse

import QKDSimkit.core as core


def get_key(alice_address: str, channel_address: str, token: str, number: int, size: int):
    """Runs handshake and starts bob procedure

    Args:
        alice_address (str): address of server
        channel_address (str): address of channel
        token (str): pre shared token
        number (int): number of keys
        size (int): size of keys (bits)
    """
    hashed = core.utils.hash_token(token)
    params = urllib.parse.urlencode({'hashed': hashed})
    conn = http.client.HTTPConnection(f"{alice_address}")
    conn.request("GET", f"/hello?{params}")
    r = conn.getresponse()
    if r.status != 200:
        return r.status
    data = r.read().decode()
    proof = core.utils.decrypt(token, data)
    conn.close()
    hash_proof = core.utils.hash_token(proof)
    params = urllib.parse.urlencode({'number': number, 'size': size, 'hashed': hashed, 'hash_proof': hash_proof})
    conn.close()
    conn1 = http.client.HTTPConnection(f"{alice_address}")
    conn1.request("GET", f"/proof?{params}")
    r = conn1.getresponse()
    if r.status == 200:
        key_list = []
        for n in range(number):
            res = core.bob.import_key(channel_address=channel_address, ID=hashed, size=size)
            if res == -1:
                print("Can't exchange a safe key")
            if isinstance(res, bytes):
                key_list.append(res.decode())
        return key_list
    else:
        return r.status


def manage_args():
    """Manages possible arguments and provides help messages"""

    parser = argparse.ArgumentParser(description='Client for Quantumacy')
    parser.add_argument('alice_address', type=str, help='Address of server/Alice [host:port]')
    parser.add_argument('channel_address', type=str, help='Address of channel [host:port]')
    parser.add_argument('-n', '--number', default=1, type=int, help="Number of keys (default: %(default)s)")
    parser.add_argument('-s', '--size', default=256, type=int, help="Size of keys (default: %(default)s)")
    return parser.parse_args()


def start_client(alice_address, channel_address, number, size):
    """Wrapper for get_key()
    Args:
        alice_address (str): address of server
        channel_address (str): address of channel
        number (int): number of keys
        size (int): size of keys (bits)
    """
    return get_key(alice_address, channel_address, '7KHuKtJ1ZsV21DknPbcsOZIXfmH1_MnKdOIGymsQ5aA=', number, size)


if __name__ == '__main__':
    args = manage_args()
    start_client(args.alice_address, args.channel_address, args.number, args.size)



# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

import setuptools
import os

thelibFolder = os.path.dirname(os.path.realpath(__file__))
requirementPath = thelibFolder + '/requirements.txt'
install_requires = []
if os.path.isfile(requirementPath):
    with open(requirementPath) as f:
        install_requires = f.read().splitlines()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="QKDSimkit",
    version="0.0.2",
    license= "MIT License",
    author="Alberto Di Meglio and Gabriele Morello",
    author_email="gabriele.morello@cern.ch",
    description="Quantum Key Distribution simulator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/CERN/Quantumacy/tree/main/QKDSimkit",
    project_urls={
        "Bug Tracker": "https://github.com/CERN/Quantumacy/issues",
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
    ],
    packages=['QKDSimkit', 'QKDSimkit.core'],
    python_requires=">=3.8",
    development_status = "3 - Alpha",
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'QKDSimkit = QKDSimkit.cli:cli',
        ],
    },
)



# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

"""This module simulates Bob's operations"""

import sys
import logging

from base64 import urlsafe_b64encode

from QKDSimkit.core.receiver import Receiver
from QKDSimkit.core.qexceptions import qsocketerror


def import_key(channel_address: str, ID: str, size: int = 256):
    """Bob's procedure to agree on a shared key

    Args:
        channel_address (str): channel address
        ID (str): identifier of alice-bob pair
        size (int): size of key in bits
    """
    channelIP, channelPort = channel_address.split(':')
    channelPort = int(channelPort)

    for count in range(0, 1000):
        bob = Receiver(ID, size)

        try:
            # connect to channel
            bob.connect_to_channel(channelIP, channelPort)
            # listen for a photon pulse on the quantum channel (this calls is blocking)
            bob.listen_quantum()
            bob.measure_photon_pulse()

        except qsocketerror as err:
            logging.error(
                "An error occurred while connecting to the quantum channel (" + str(err) + "). Disconnecting.")
            sys.exit()
        except Exception as err:
            logging.error("An error occurred (" + str(err) + "). Disconnecting.")
            sys.exit()

        # now connect to classic channel
        try:
            # connect to channel
            bob.reset_socket()
            bob.connect_to_channel(channelIP, channelPort)
        except qsocketerror as err:
            logging.error(
                "An error occurred while connecting to the classic channel (" + str(err) + "). Disconnecting.")
            sys.exit()
        except Exception as err:
            logging.error("An error occurred (" + str(err) + "). Disconnecting.")
            sys.exit()

        # exchange basis
        try:
            # send Bob's chosen basis
            bob.send('bob-other_bases', repr(bob.bases))
            # listen for Alice's reconciled key through classic channel
            bob.listen_for('alice', 'reconciled_key')

        except qsocketerror as err:
            logging.error(
                "An error occurred while connecting to the classic channel (" + str(err) + "). Disconnecting.")
            sys.exit()
        except Exception as err:
            logging.error("An error occurred (" + str(err) + "). Disconnecting.")
            sys.exit()

        # create key and public sub key
        bob.create_keys()

        # exchange sub key
        try:
            # listen for Alice's sub key
            bob.listen_for('alice', 'other_sub_key')

            # send the public sub key
            bob.send('bob-other_sub_key', repr(bob.sub_shared_key))

            bob.decision = bob.validate()

            # listen for Alice's sub key
            bob.listen_for('alice', 'other_decision')

            # send decision
            bob.send('bob-other_decision', repr(bob.decision))

        except qsocketerror as err:
            logging.error(
                "An error occurred while connecting to the classic channel (" + str(err) + "). Disconnecting.")
            sys.exit()
        except Exception as err:
            logging.error("An error occurred (" + str(err) + "). Disconnecting.")
            sys.exit()

        bob.reset_socket()

        # choose what to do
        if bob.decision == bob.other_decision and bob.decision == 1:
            # return a correct key
            bob.get_key()
            logging.info("Success!")
            bob.key = bob.key[:size]
            bob.key = [int("".join(map(str, bob.key[i:i + 8])), 2) for i in range(0, len(bob.key), 8)]
            return urlsafe_b64encode(bytearray(bob.key))
        elif bob.decision == bob.other_decision and bob.decision == 0:
            # retry
            logging.info("Failed to match key, trying again")
            continue
        else:
            # exit
            logging.warning("Failed! Noise or eavesdropper detected")
            return -1
    logging.warning("Error: too many attempts to find a shared key")
    return -1


if __name__ == '__main__':
    import_key('127.0.0.1:8000', 'id', 256)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

"""This module contains methods to operate a p2p node"""

import argparse
import asyncio
import logging
import json

import uvicorn

from aiocache import Cache
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

import QKDSimkit.core as core


logging.basicConfig(level=logging.ERROR)

cache = Cache(Cache.REDIS, endpoint="localhost", port=6379, namespace="p2p_server")

alice_app = FastAPI()
bob_app = FastAPI()


async def answer_get(number: int, size: int, ID: str, type: str):
    """Starts a node

    Args:
        number (int): number of keys
        size (int): size of keys (bits)
        ID (str): identifier of a pair of nodes
        type (str): alice or bob (sender or receiver)
    Returns:
        keys
    """
    answer = {}
    keys = []
    address = await cache.get("channel_address")
    for i in range(number):
        if type == 'Alice':
            key = core.alice.import_key(channel_address=address, ID=ID, size=size)
        if type == 'Bob':
            key = core.bob.import_key(channel_address=address, ID=ID, size=size)
        keys.append({"key_ID": i, "key": key})
    answer["keys"] = keys
    return answer


@alice_app.get("/test")
async def root(number: int = 1, size: int = 256, ID: str = 'id'):
    """http request handler for alice

        Args:
            number (int): number of keys
            size (int): size of keys (bits)
            ID (str): identifier of a pair of nodes
        Returns:
            keys
    """
    return await answer_get(number, size, ID, 'Alice')


@bob_app.get("/test")
async def root(number: int = 1, size: int = 256, ID: str = 'id'):
    """http request handler for bob

        Args:
            number (int): number of keys
            size (int): size of keys (bits)
            ID (str): identifier of a pair of nodes
        Returns:
            keys
    """
    return await answer_get(number, size, ID, 'Bob')


origins = [
    "*"
]

alice_app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

bob_app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


def manage_args():
    """Manages possible arguments and provides help messages"""

    parser = argparse.ArgumentParser(description='Server for Quantumacy')
    parser.add_argument('node', choices=['alice', 'bob'],
                       help='Choose how to run this node')
    parser.add_argument('-c', '--channel_address', default=':5000', type=str,
                        help='Specify the address of the channel [host:port]')
    parser.add_argument('-a', '--address', default='127.0.0.1:5003', type=str,
                        help='Bind socket to this address (default: %(default)s)')
    return parser.parse_args()


def start_p2p(node: str, address: str, channel_address: str):
    """Starts server

    Args:
        node (str): type of node [alice, bob]
        address (str): where to bind socket
    """
    asyncio.run(cache.set('channel_address', channel_address))
    uvicorn.run('QKDSimkit.p2p_servers:{}_app'.format(node), host=address.split(':')[0], port=int(address.split(':')[1]))


if __name__ == '__main__':
    args = manage_args()
    start_p2p(args.node, args.address)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

"""This module contains channel's interface"""

import argparse

from QKDSimkit.core import channel


def start_channel(address: str, noise: float, eve: bool):
    """Starts channel
    Args:
        address (str): where to bind the channel
        noise (float): ratio of noise in channel
        eve (bool): simulate an eavesdropper in channel
    """
    # instantiate a receiver channel
    theChannel = channel.public_channel(address, noise, eve)

    # initiate the channel and listen for connections
    theChannel.initiate_channel()


def manage_args():
    """Manages possible arguments and provides help messages"""

    parser = argparse.ArgumentParser(description='Channel for Quantumacy')
    parser.add_argument('-a', '--address', default=':8000', type=str,
                        help='Bind socket to this address (default: %(default)s)')
    parser.add_argument('-n', '--noise', default=0.0, type=float,
                        help='Set a noise value for channel, type a float number in [0,1] (default: %(default)s)')
    parser.add_argument('-e', '--eve', action='store_true', help='Add an eavesdropper to the channel')
    return parser


if __name__ == '__main__':
    args = manage_args().parse_args()
    start_channel(args.address, args.noise, args.eve)



# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.


class qsocketerror(Exception):
    pass

class qobjecterror(Exception):
    pass



# -*- coding: utf-8 -*-
# This code is part of QKDSimkit.
#
# SPDX-License-Identifier: MIT
#
# (C) Copyright 2021 CERN.

import abc
import ast
import logging
import re
import select
import socket

from .qexceptions import qsocketerror
from .utils import validate

MIN_SHARED = 20
BUFFER_SIZE = 8192
TIMEOUT_IN_SECONDS = 1
CONNECTION_ATTEMPTS = 50
MAX_REPETITIONS = 1000
MIN_SHARED_PERCENT = 0.89


class Node(object):
    """Father class for receiver and sender

    Args:
        ID (str): identifier of alice-bob pair
        size (int): size of key in bits

    Attributes:
        min_shared (int): 
        buffer_size (int):
        timeout_in_seconds (int):
        connection_attempts (int):
        max_repetitions (int):
        min_shared_percent (float):
        socket: socket
        ID (str): identifier of alice-bob pair
        photon_pulse (list): list of photons
        bases (list): list of bases
        other_bases (list): list of bases of the other node
        reconciled_key (list): key with only common values and ""
        shared_key (list): key with only common values
        sub_shared_key (list): first half of the key
        other_sub_key (list): first half of the key of the other node
        decision (int): result of comparison between shared part of the key
        other_decision (int): result of comparison between ke ys
        not_shared_key (list):
        key (list): non-revealed part of the key
        fragments (list):
        regex: utility regex
        photon_pulse_size (int): number of photons exchanged photons
    """
    def __init__(self, ID, size):
        self.min_shared = MIN_SHARED
        self.buffer_size = BUFFER_SIZE
        self.timeout_in_seconds = TIMEOUT_IN_SECONDS
        self.connection_attempts = CONNECTION_ATTEMPTS
        self.max_repetitions = MAX_REPETITIONS
        self.min_shared_percent = MIN_SHARED_PERCENT
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.ID = ID
        self.photon_pulse = []
        self.bases = []
        self.other_bases = []
        self.reconciled_key = []
        self.shared_key = []
        self.sub_shared_key = []
        self.other_sub_key = []
        self.decision = 0
        self.other_decision = 0
        self.key = []
        self.regex = r'\((.)*\):'
        self.photon_pulse_size = size * 5

    def connect_to_channel(self, address: str, port: int):
        """It starts the connection with the channel

        Args:
            address (str): address
            port (int): port

        """
        try:
            self.socket.connect((address, port))
        except socket.error:
            raise qsocketerror("unable to connect")

    def reset_socket(self):
        """Reset the previously started socket
        """
        try:
            self.socket.close()
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        except Exception as e:
            logging.info("Failed to reset socket:\n" + str(e))
            raise ConnectionError

    def ownMessage(self, addr: str) -> bool:
        """Check if a message comes from this node

        Args:
            addr (str): address of the received message
        """
        hostname = socket.gethostname()
        local_ip = socket.gethostbyname(hostname)
        if local_ip == addr:
            return True
        return False

    def create_keys(self):
        """It calls two functions to elaborate th photon pulse in actual keys"""
        self.create_shared_key()
        self.create_sub_shared_key()

    def create_shared_key(self):
        """Converts the whole photon pulse in a list of bit"""
        for i in range(len(self.photon_pulse)):
            if self.reconciled_key[i] != "":
                self.shared_key.append(self.photon_pulse[i].bit)

    def create_sub_shared_key(self):
        """Create the part of the key that will be sent to the other node (first half)"""
        self.sub_shared_key = self.shared_key[:(len(self.shared_key) // 2)]

    def get_key(self):
        """Create the part of the key that will not be sent and it will be used as symmetric key (second half)"""
        self.key = self.shared_key[(len(self.shared_key) // 2):]

    def listen_for(self, sender: str, attr: str):
        """Receive a message and store it in the right place

        Args:
            sender (str): name of the node from which we expect to receive
            attr (str): name of the attribute that will store the content of the received message
            """
        try:
            logging.info("Listening to classical channel for " + attr)
            while True:
                message = self.recv(sender + '-' + attr)
                try:
                    literal = ast.literal_eval(message)
                except ValueError as VE:
                    logging.error("Value Error: " + str(VE))
                    pass
                else:
                    setattr(self, attr, literal)
                    break
        except socket.error:
            raise qsocketerror("not connected to any channel")

    def validate(self) -> int:
        """Wrapper of utils.validate it manages different outputs, it also gives info of eventual errors

        Returns:
            int: 1: keys are equals, 0: error rate is below a given percent, -1: error rate too high
        """
        percent = validate(self.sub_shared_key, self.other_sub_key)
        logging.info('Correct bits percentage: ' + str(percent))
        if percent == 1:
            return 1
        if self.min_shared_percent <= percent < 1:
            return 0
        if percent < self.min_shared_percent:
            return -1

    def recv_all(self) -> list:
        """receive a message
        receive from socket, sum every part of the message, check if the message is commplete, check if the ID of the
        sender corresponds to the ID of the receiver, return a list whit header and payload, please note that the ID
        is not returned

        Returns:
            fragments (list): list made by header and payload
        """
        message = ''
        ready = select.select([self.socket], [], [], self.timeout_in_seconds)
        if ready[0]:
            while True:
                data_recv = self.socket.recv(self.buffer_size).decode()
                message += re.sub(self.regex, '', data_recv)  # removing ('xxx.xxx.xxx.xxx', xxxxx):
                if message.count(':') >= 3:  # checking if payload started and finished example: 'ID:head:payload:'
                    fragments = message.split(':', 3)
                    if fragments[0] != self.ID:  # this message doesn't belong to this node
                        message = fragments[3]  # we can discard it
                        continue
                    else:
                        return fragments[1:]  # we don't need to return ID because we already checked it is correct

    @abc.abstractmethod
    def send(self, header: str, message: str):
        """abstract method"""
        print("send(): Override me")

    @abc.abstractmethod
    def recv(self, header: str):
        """abstract method"""
        print("receive(): Override me")
